from cg.utils.enums import StrEnum


class CustomerNames(StrEnum):
    cust999: str = "cust999"
    cust032: str = "cust032"
    cust001: str = "cust001"
    cust132: str = "cust132"


class CostCenters(StrEnum):
    ki: str = "ki"
    kth: str = "kth"
